REM
REM Package of test procedures for mod_owa
REM

create or replace package MODOWA_TEST_PKG is
  --
  type VCTAB is table of varchar2(32767) index by binary_integer;
  --
  function AUTHORIZE return boolean;
  --
  procedure CHECK_CACHE(URI out varchar2, CHKSUM out varchar2,
                        MTYPE out varchar2, LIFE out number);
  --
  procedure TEST_ARGUMENTS(N in number default 300, A in varchar2);
  --
  procedure TEST_ARRAY(N in number, A in VCTAB);
  --
  procedure TEST_ARRAY(N in number, A in OWA_ARRAY);
  --
  procedure TEST_LARGE(NUM_ENTRIES in number, NAME_ARRAY in VCTAB,
                       VALUE_ARRAY in VCTAB, RESERVED in VCTAB);
  --
  procedure TEST_FLEX(NUM_ENTRIES in number, NAME_ARRAY in VCTAB,
                      VALUE_ARRAY in VCTAB, RESERVED in VCTAB);
  --
  procedure TEST_FLEX(NUM_ENTRIES in number, NAME_ARRAY in OWA_ARRAY,
                      VALUE_ARRAY in OWA_ARRAY, RESERVED in OWA_ARRAY);
  --
  procedure TEST_COOKIE(A in varchar2);
  --
  procedure TEST_AUTH;
  --
  procedure TEST_STATUS(ERRCODE in varchar2 default null);
  --
  procedure TEST_RAWARG(R in raw);
  --
  procedure TEST_RAWBUF(R in raw);
  --
  procedure TEST_PLAIN;
  --
  procedure TEST_IMAGE;
  --
  procedure TEST_XML;
  --
  procedure TEST_TILDE(R in raw, MTYPE in out varchar2, PBLOB out blob);
  --
  procedure TEST_NCHAR(A in nvarchar2);
  --
  procedure TEST_REDIRECT(U in varchar2);
  --
  procedure TEST_NOHEADER(U in varchar2);
  --
  procedure TEST_SCRIPTNAME;
  --
  procedure EUROCHARS(N in number default 10);
  --
  procedure CHINESE(N in number default 10);
  --
end MODOWA_TEST_PKG;
/

create or replace package body MODOWA_TEST_PKG is
  --
  RETRY_FLAG  boolean := false;
  CACHE_URI   varchar2(2000) := null;
  CACHE_SUM   varchar2(2000) := null;
  --
  function AUTHORIZE return boolean is
  begin
    return(TRUE);
  end AUTHORIZE;
  --
  procedure CHECK_CACHE(URI out varchar2, CHKSUM out varchar2,
                        MTYPE out varchar2, LIFE out number) is
  begin
    URI := CACHE_URI;
    CHKSUM := CACHE_SUM;
    LIFE := 600; -- 10 minutes
    MTYPE := 'text/html';
    CACHE_URI := '';
    CACHE_SUM := '';
  end CHECK_CACHE;
  --
  function ESCAPE_URL_ARG(C in varchar2) return varchar2 is
    BUFFER varchar2(2000);
    BCHAR  varchar2(10);
    SLEN   number;
    ALPHA  varchar2(64) :=
           '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_-';
  begin
    if (C is null) then
      SLEN := 0;
    else
      SLEN := length(C);
    end if;
    BUFFER := '';
    for I in 1..SLEN loop
      BCHAR := substr(C, I, 1);
      if (BCHAR = ' ') then
        BCHAR := '+';
      elsif (instr(ALPHA, BCHAR) = 0) then
        BCHAR := '%'||RAWTOHEX(UTL_RAW.CAST_TO_RAW(BCHAR));
      end if;
      BUFFER := BUFFER||BCHAR;
    end loop;
    return(BUFFER);
  end ESCAPE_URL_ARG;
  --
  procedure TEST_ARGUMENTS(N in number default 300, A in varchar2) is
    CACHE_FLAG varchar2(30);
  begin
    CACHE_FLAG := OWA_UTIL.GET_CGI_ENV('MODOWA_PAGE_CACHE');
    if (lower(CACHE_FLAG) = 'enabled') then
      CACHE_URI := '/docs/owapages/testargs_'||N||'_.htm';
      CACHE_SUM := A;
    else
      RETRY_FLAG := true;
    end if;
    if (RETRY_FLAG) then
      OWA_UTIL.MIME_HEADER('text/html', TRUE);
      HTP.P('<html><body bgcolor="#ddddee" text="#000000"><p>');
      for I in 1..N loop
        HTP.P(to_char(I,'9999')||': '||A||'<br>');
      end loop;
      HTP.P('</p></body></html>');
      RETRY_FLAG := false;
    else
      RETRY_FLAG := true;
    end if;
  end TEST_ARGUMENTS;
  --
  procedure TEST_ARRAY(N in number, A in VCTAB) is
  begin
    OWA_UTIL.MIME_HEADER('text/html', TRUE);
    HTP.P('<html><body bgcolor="#ddddee" text="#000000"><p>');
    for I in 1..N loop
      HTP.P(to_char(I)||': '||A(I)||'<br>');
    end loop;
    HTP.P('</p></body></html>');
  end TEST_ARRAY;
  --
  procedure TEST_ARRAY(N in number, A in OWA_ARRAY) is
    A_NEW VCTAB;
  begin
    for I in 1..N loop
      A_NEW(I) := A(I);
    end loop;
    TEST_ARRAY(N, A_NEW);
  end TEST_ARRAY;
  --
  procedure TEST_LARGE(NUM_ENTRIES in number, NAME_ARRAY in VCTAB,
                       VALUE_ARRAY in VCTAB, RESERVED in VCTAB) is
  begin
    OWA_UTIL.MIME_HEADER('text/html', TRUE);
    HTP.P('<html><body bgcolor="#ddddee" text="#000000"><pre>');
    for I in 1..NUM_ENTRIES loop
      HTP.P(NAME_ARRAY(I)||' ['||lengthb(VALUE_ARRAY(I))||']');
      HTP.P(VALUE_ARRAY(I));
    end loop;
    HTP.P('</pre></body></html>');
  end;
  --
  procedure TEST_FLEX(NUM_ENTRIES in number, NAME_ARRAY in VCTAB,
                      VALUE_ARRAY in VCTAB, RESERVED in VCTAB) is
  begin
    OWA_UTIL.MIME_HEADER('text/html', TRUE);
    HTP.P('<html><body bgcolor="#ddddee" text="#000000"><p>');
    for I in 1..NUM_ENTRIES loop
      HTP.P(NAME_ARRAY(I)||'='||ESCAPE_URL_ARG(VALUE_ARRAY(I))||'<br>');
    end loop;
    HTP.P('</p></body></html>');
  end;
  --
  procedure TEST_FLEX(NUM_ENTRIES in number, NAME_ARRAY in OWA_ARRAY,
                      VALUE_ARRAY in OWA_ARRAY, RESERVED in OWA_ARRAY) is
    NAME_NEW     VCTAB;
    VALUE_NEW    VCTAB;
    RESERVED_NEW VCTAB;
  begin
    for I in 1..NUM_ENTRIES loop
      NAME_NEW(I) := NAME_ARRAY(I);
      VALUE_NEW(I) := VALUE_ARRAY(I);
      RESERVED_NEW(I) := RESERVED(I);
    end loop;
    TEST_FLEX(NUM_ENTRIES, NAME_NEW, VALUE_NEW, RESERVED_NEW);
  end;
  --
  procedure TEST_COOKIE(A in varchar2) is
    ACOOKIE owa_cookie.cookie;
  begin
    ACOOKIE := OWA_COOKIE.GET('modowa_cookie');
    OWA_UTIL.MIME_HEADER('text/html', FALSE);
    OWA_COOKIE.SEND('modowa_cookie', 'value_'||A);
    OWA_UTIL.HTTP_HEADER_CLOSE;
    HTP.P('<html><body bgcolor="#ddddee" text="#000000"><p>');
    if (ACOOKIE.NUM_VALS > 0) then
      for I in 1..ACOOKIE.NUM_VALS loop
        HTP.P('modowa_cookie : ['||ACOOKIE.VALS(I)||']<br>');
      end loop;
    else
      HTP.P('modowa_cookie not found');
    end if;
    HTP.P('</p><p><b>Set</b> modowa_cookie : [value_'||A||']</p>');
    HTP.P('</p></body></html>');
  end TEST_COOKIE;
  --
  procedure TEST_AUTH is
    AUTHSTRING varchar2(250);
  begin
    AUTHSTRING := OWA_UTIL.GET_CGI_ENV('HTTP_AUTHORIZATION');
    if (AUTHSTRING is null) then
      OWA_UTIL.MIME_HEADER('text/html', FALSE);
      HTP.PRN('WWW-Authenticate: Basic realm="modowa"');
      OWA_UTIL.HTTP_HEADER_CLOSE;
      return;
    end if;
    OWA_UTIL.MIME_HEADER('text/html', TRUE);
    HTP.P('<html><body bgcolor="#ddddee" text="#000000"><p>');
    HTP.P(AUTHSTRING||'<br>');
    HTP.P(OWA_SEC.GET_USER_ID()||'/'||OWA_SEC.GET_PASSWORD());
    HTP.P('</p></body></html>');
  end TEST_AUTH;
  --
  procedure TEST_STATUS(ERRCODE in varchar2 default null) is
    ECODE number;
  begin
    if (ERRCODE is not null) then
      ECODE := to_number(ERRCODE);
    else
      ECODE := 500;
    end if;
    OWA_UTIL.STATUS_LINE(ECODE, null, TRUE);
  end TEST_STATUS;
  --
  procedure TEST_RAWARG(R in RAW) is
  begin
    OWA_UTIL.MIME_HEADER('text/plain', TRUE);
    HTP.P('You passed in '||UTL_RAW.CAST_TO_VARCHAR2(R));
  end TEST_RAWARG;
  --
  procedure TEST_RAWBUF(R in RAW) is
  begin
    OWA_UTIL.MIME_HEADER('text/plain', TRUE);
    HTP.P(RAWTOHEX(R));
  end TEST_RAWBUF;
  --
  procedure TEST_PLAIN is
  begin
    OWA_UTIL.MIME_HEADER('text/plain', TRUE);
    HTP.P('The system time is '||to_char(sysdate,'YYYY/MM/DD HH24:MI:SS'));
  end TEST_PLAIN;
  --
  procedure TEST_IMAGE is
  begin
    OWA_UTIL.MIME_HEADER('image/gif', TRUE);
    HTP.PRN(UTL_RAW.CAST_TO_VARCHAR2(
            hextoraw('47494638396114001600C20000FFFFFF'||
                     'FF9999FF3333CCFFFF66000000000000'||
                     '000000000021FE4E5468697320617274'||
                     '20697320696E20746865207075626C69'||
                     '6320646F6D61696E2E204B6576696E20'||
                     '4875676865732C206B6576696E684065'||
                     '69742E636F6D2C2053657074656D6265'||
                     '7220313939350021F90401000003002C'||
                     '000000001400160000034038BADCFE30'||
                     'CA49AB5D25975B82171BD50181201061'||
                     'D4056479A64F21942561C34E379B2FDE'||
                     'C83C1BCAC700124E4362518842550A49'||
                     '654C73A95AAFD8AC569100003B')));
  end TEST_IMAGE;
  --
  procedure TEST_XML is
  begin
    OWA_UTIL.MIME_HEADER('text/xml', TRUE);
    HTP.P('<anmial><name>Cat</name><sound>Meow</sound></animal>');
  end TEST_XML;
  --
  procedure TEST_TILDE(R in raw, MTYPE in out varchar2, PBLOB out blob) is
    BUFFER raw(256);
    RLEN   number;
  begin
    RLEN := UTL_RAW.LENGTH(R);
    if (RLEN > 0) then
      MTYPE := 'text/plain';
      DBMS_LOB.CREATETEMPORARY(PBLOB, FALSE);
      DBMS_LOB.OPEN(PBLOB, DBMS_LOB.LOB_READWRITE);
      DBMS_LOB.WRITE(PBLOB, RLEN, 1, R);
      DBMS_LOB.CLOSE(PBLOB);
    else
      MTYPE := 'image/gif';
      BUFFER := hextoraw('47494638396114001600C20000FFFFFF'||
                         'FF9999FF3333CCFFFF66000000000000'||
                         '000000000021FE4E5468697320617274'||
                         '20697320696E20746865207075626C69'||
                         '6320646F6D61696E2E204B6576696E20'||
                         '4875676865732C206B6576696E684065'||
                         '69742E636F6D2C2053657074656D6265'||
                         '7220313939350021F90401000003002C'||
                         '000000001400160000034038BADCFE30'||
                         'CA49AB5D25975B82171BD50181201061'||
                         'D4056479A64F21942561C34E379B2FDE'||
                         'C83C1BCAC700124E4362518842550A49'||
                         '654C73A95AAFD8AC569100003B');
      DBMS_LOB.CREATETEMPORARY(PBLOB, FALSE);
      DBMS_LOB.OPEN(PBLOB, DBMS_LOB.LOB_READWRITE);
      DBMS_LOB.WRITE(PBLOB, UTL_RAW.LENGTH(BUFFER), 1, BUFFER);
      DBMS_LOB.CLOSE(PBLOB);
    end if;
  end TEST_TILDE;
  --
  procedure TEST_NCHAR(A in nvarchar2) is
  begin
    OWA_UTIL.MIME_HEADER('text/html', TRUE);
    HTP.P('<html><body bgcolor="white" text="black">');
    HTP.P('<p>Length of NCHAR argument is '||to_char(length(A))||'.</p>');
    HTP.P('</body></html>');
  end TEST_NCHAR;
  --
  procedure TEST_REDIRECT(U in varchar2) is
  begin
    OWA_UTIL.REDIRECT_URL(U);
  end TEST_REDIRECT;
  --
  procedure TEST_NOHEADER(U in varchar2) is
  begin
    HTP.P('<html><body bgcolor=#ffffc0 text=#000000>A');
    for I in 1..60 loop
      HTP.P(U);
    end loop;
    HTP.P('Z</body></html>');
  end TEST_NOHEADER;
  --
  procedure TEST_SCRIPTNAME is
    SCRIPT   varchar2(200);
    PROCNAME varchar2(200);
    IDX      number;
  begin
    OWA_UTIL.MIME_HEADER('text/html', TRUE);
    HTP.P('<html><body bgcolor="#ddddee" text="#000000"><p>');
    SCRIPT := OWA_UTIL.GET_OWA_SERVICE_PATH;
    PROCNAME := OWA_UTIL.GET_PROCEDURE;
    HTP.P('native ['||SCRIPT||'] ['||PROCNAME||']<br>');
    SCRIPT := OWA_UTIL.GET_CGI_ENV('SCRIPT_NAME');
    PROCNAME := OWA_UTIL.GET_CGI_ENV('PATH_INFO');
    IDX := length(SCRIPT);
    HTP.P('original ['||SCRIPT||'] ['||PROCNAME||']<br>');
    if (substr(SCRIPT, IDX, 1) = '/') then
      SCRIPT := substr(SCRIPT, 1, IDX-1);
    end if;
    if (substr(PROCNAME,1,1) <> '/') then
      PROCNAME := '/'||PROCNAME;
    end if;
    HTP.P('adjusted ['||SCRIPT||'] ['||PROCNAME||']<br>');
    while (IDX > 0) loop
      IDX := instr(PROCNAME, '/');
      if (IDX > 0) then -- need to adjust
        SCRIPT := SCRIPT||substr(PROCNAME, 1, IDX);
        PROCNAME := substr(PROCNAME, IDX + 1);
        HTP.P('rewrite ['||SCRIPT||'] ['||PROCNAME||']<br>');
      end if;
    end loop;
    HTP.P('</p></body></html>');
  end TEST_SCRIPTNAME;
  --
  procedure EUROCHARS(N in number default 10) is
    BUF varchar2(30);
  begin
    HTP.P('<html><body bgcolor="#ffffff" text="#000000"><p>');
    BUF :='C'||chr(228)||'t D'||chr(246)||'g<br>';
    for I in 1..N loop
      HTP.P(BUF);
    end loop;
    HTP.P('</p></body></html>');
  end EUROCHARS;
  --
  procedure CHINESE(N in number default 10) is
    BUF varchar2(30);
  begin
    HTP.P('<html><body bgcolor="#ffffff" text="#000000"><p>');
    for I in 1..N loop
      HTP.P(i||':'||chr(15113393)||chr(15106235)||chr(15175814)||'<br>');
    end loop;
    HTP.P('</p></body></html>');
  end CHINESE;
  --
end MODOWA_TEST_PKG;
/
